from abc import ABC, abstractmethod
import json
import tomllib
import configparser
from pathlib import Path

from .config import BuildStep


class ParserInterface(ABC):
    @abstractmethod
    def parse(self, project_root: Path) -> BuildStep | None:
        pass


def chain_of_command(
    project_root: Path, *parsers: ParserInterface
) -> list[BuildStep]:
    all_build_steps = (parser.parse(project_root) for parser in parsers)
    return [s for s in all_build_steps if s is not None]  # type: ignore


class ToxParser(ParserInterface):
    def parse(self, project_root: Path) -> BuildStep | None:
        tox_ini = project_root / 'tox.ini'
        if tox_ini.exists():
            return BuildStep(cmd='tox')

        pyproject_toml = project_root / 'pyproject.toml'
        if pyproject_toml.exists():
            config = tomllib.loads(pyproject_toml.read_text())
            if config['tool']['tox']['legacy_tox_ini']:
                return BuildStep(cmd='tox')

        setup_cfg = project_root / 'setup.cfg'
        if setup_cfg.exists():
            parser = configparser.ConfigParser()
            parser.read([str(setup_cfg)])
            print(parser)
            if parser.has_section('tox:tox'):
                return BuildStep(cmd='tox')

        return None


class PyProjectParser(ParserInterface):
    def parse(self, project_root: Path) -> BuildStep | None:
        pyproject_toml = project_root / 'pyproject.toml'
        if pyproject_toml.exists():
            config = tomllib.loads(pyproject_toml.read_text())
            build_backend = config.get('build-system', {}).get('build-backend')
            if not build_backend:
                return None

            if build_backend == 'setuptools.build_meta':
                return BuildStep(cmd='python -m build')
            elif build_backend in ['flit.buildapi', 'flit_core.buildapi']:
                # NOTE: flit could also do upload
                return BuildStep(cmd='flit build')
            elif build_backend == 'poetry.core.masonry.api':
                return BuildStep(cmd='poetry build')

        return None


class PackageJsonParser(ParserInterface):
    section: str

    def __init__(self, section: str):
        self.section = section

    def parse(self, project_root: Path) -> BuildStep | None:
        package_json = project_root / 'package.json'
        if package_json.exists():
            config = json.loads(package_json.read_text())
            if config['scripts'].get(self.section):
                return BuildStep(cmd=f'npm run {self.section}')
        return None


class DockerParser(ParserInterface):
    def parse(self, project_root: Path) -> BuildStep | None:
        dockerfile = project_root / 'Dockerfile'
        dockerfiles = list(project_root.glob('Dockerfile*'))
        if dockerfile.exists():
            if len(dockerfiles) == 1:
                return BuildStep(cmd='docker buildx build .')
            else:
                return BuildStep(cmd='docker buildx build -F Dockerfile .')
        elif len(dockerfiles) == 1:
            return BuildStep(
                cmd=f'docker buildx build -F {dockerfiles[0].name} .'
            )

        return None



#!/usr/bin/env python
import typer
from pathlib import Path
from io import StringIO
import yaml
from .config import Config, Component
from .state import state_update
from .build import build_component
from . import parsers

app = typer.Typer()


@app.callback()
def main():
    # Set global options here
    pass


@app.command()
def build(config: str = 'abuild.yaml'):
    """Build components listed in the config file"""
    cfg = Config.from_file(Path(config))
    for component in cfg.components:
        with state_update(component.path, cfg.state_file) as need_rebuild:
            if need_rebuild:
                print(f'Building component: {component.display_name}')
                build_component(component)


@app.command()
def parse():
    """Parse local directory and print a suggested configuration"""
    config = Config(components=[])
    for subdir in sorted(Path('.').iterdir()):
        if subdir.is_dir():
            steps = parsers.chain_of_command(
                subdir,
                parsers.ToxParser(),
                parsers.PyProjectParser(),
                parsers.DockerParser(),
                parsers.PackageJsonParser('test'),
                parsers.PackageJsonParser('build'),
            )
            if len(steps):
                config.components.append(
                    Component(path=str(subdir), steps=steps)
                )
    buf = StringIO()
    yaml.dump(config.model_dump(mode='json', exclude_unset=True), buf)
    print(buf.getvalue())


if __name__ == '__main__':
    app.run()



from pathlib import Path
import subprocess
from .config import Component, BuildStep


class BuildError(Exception):
    pass


def build_component(component: Component):
    for step in component.steps:
        print(f'Running step: {step.display_name}')
        run_build_step(step, cwd=component.path)


def run_build_step(step: BuildStep, cwd: Path):
    # TODO: Small window with process output
    proc = subprocess.run(
        step.cmd,
        shell=True,
        cwd=str(cwd),
        stdout=subprocess.PIPE,
        stderr=subprocess.STDOUT,
    )
    print('Output:')
    print(proc.stdout.decode('utf-8'))
    if proc.returncode and step.break_on_error:
        # TODO: Make full process output visible
        raise BuildError(
            f'Step {step.display_name} failed with return code {proc.returncode}'
        )



from typing import cast
from pathlib import Path
import fnmatch
from abc import ABC, abstractmethod


class FileSkipperInterface(ABC):
    @abstractmethod
    def skip(self, path: Path) -> bool:
        pass

    @abstractmethod
    def update(self, ignorefile: Path) -> 'FileSkipperInterface':
        pass


class NeverSkip(FileSkipperInterface):
    def skip(self, path: Path) -> bool:
        return False

    def update(self, ignorefile: Path) -> 'NeverSkip':
        return NeverSkip()


EmptyList = cast(list[str], object())


class GlobSkipper(FileSkipperInterface):
    patterns: list[str]

    def __init__(self, patterns: list[str] = EmptyList):
        if patterns is EmptyList:
            self.patterns = []
        else:
            self.patterns = patterns

    def skip(self, path: Path) -> bool:
        for pattern in self.patterns:
            if fnmatch.fnmatchcase(str(path), pattern):
                return True
            print(f'No match: {pattern} -- {str(path)}')
        return False

    def update(self, ignorefile: Path) -> 'GlobSkipper':
        new_patterns = [
            line.strip('\n') for line in ignorefile.read_text().splitlines()
        ]
        return GlobSkipper(self.patterns + new_patterns)

    def __repr__(self) -> str:
        return f'GlobSkipper({self.patterns})'

    def __str__(self) -> str:
        return repr(self)



from typing import Iterator
import json
from contextlib import contextmanager
from datetime import datetime
from pathlib import Path
from pydantic import BaseModel, Field

from .hashing import hash_directory


class ComponentState(BaseModel):
    name: Path
    hash: str


class FullState(BaseModel):
    components: list[ComponentState] = Field(default_factory=list)
    timestamp: datetime = Field(default_factory=datetime.utcnow)

    @classmethod
    def from_file(cls, filepath: Path):
        if not filepath.exists():
            return cls()
        with filepath.open('r') as f:
            return cls(**json.load(f))

    def to_file(self, filepath: Path):
        with filepath.open('w') as f:
            f.write(self.model_dump_json())


@contextmanager
def state_update(component: Path, state_file: Path) -> Iterator[bool]:
    state = FullState.from_file(state_file)
    idx = [i for i, c in enumerate(state.components) if c.name == component]
    current_hash = hash_directory(component)

    if len(idx) == 0:
        # not existing
        yield True
        FullState(
            components=state.components
            + [ComponentState(name=component, hash=current_hash)]
        ).to_file(state_file)
    else:
        full_component = state.components.pop(idx[0])
        if current_hash == full_component.hash:
            yield False
        else:
            yield True
            FullState(
                components=state.components
                + [ComponentState(name=component, hash=current_hash)]
            ).to_file(state_file)



import pytest
from pathlib import Path
from abuild import hashing
from abuild import ignore


def describe_all_files():
    @pytest.fixture
    def populated_dir(tmp_path: Path):
        (tmp_path / 'a_file.txt').write_text('a_file')
        (tmp_path / 'another_file.txt').write_text('another_file')
        (tmp_path / '.abuildignore').write_text('another_*')
        yield tmp_path

    def test_should_give_a_list_of_files(tmp_path: Path):
        assert sorted(hashing.all_files(tmp_path, hashing.NeverSkip())) == []

    def test_should_list_files_present(populated_dir: Path):
        assert list(
            map(
                str,
                sorted(hashing.all_files(populated_dir, hashing.NeverSkip())),
            )
        ) == [
            f'{str(populated_dir)}/.abuildignore',
            f'{str(populated_dir)}/a_file.txt',
            f'{str(populated_dir)}/another_file.txt',
        ]

    def test_should_drop_files_that_match_ignore(populated_dir: Path):
        assert list(
            map(
                str,
                sorted(hashing.all_files(populated_dir, ignore.GlobSkipper())),
            )
        ) == [
            f'{str(populated_dir)}/.abuildignore',
            f'{str(populated_dir)}/a_file.txt',
        ]



import pytest
from pathlib import Path
import json
from abuild.config import BuildStep
from abuild import parsers


def describe_tox_parser():
    def test_should_see_tox_ini(tmp_path: Path):
        (tmp_path / 'tox.ini').write_text('[tox]')
        assert parsers.ToxParser().parse(tmp_path) == BuildStep(cmd='tox')

    def test_should_parse_pyproject_toml(tmp_path: Path):
        (tmp_path / 'pyproject.toml').write_text(
            '\n'.join(['[tool.tox]', 'legacy_tox_ini = """[tox]"""'])
        )
        assert parsers.ToxParser().parse(tmp_path) == BuildStep(cmd='tox')

    def test_should_parse_setup_cfg(tmp_path: Path):
        (tmp_path / 'setup.cfg').write_text(
            '\n'.join(
                [
                    '[tox:tox]',
                    'min_version = 4.0',
                    '',
                    '[testenv]',
                    'deps = pytest',
                ]
            )
        )
        assert parsers.ToxParser().parse(tmp_path) == BuildStep(cmd='tox')

    def test_should_not_return_on_empty_dir(tmp_path: Path):
        assert parsers.ToxParser().parse(tmp_path) is None


def describe_pyproject_parser():
    def test_should_parse_setuptools(tmp_path: Path):
        (tmp_path / 'pyproject.toml').write_text(
            '\n'.join(
                ['[build-system]', 'build-backend = "setuptools.build_meta"']
            )
        )
        assert parsers.PyProjectParser().parse(tmp_path) == BuildStep(
            cmd='python -m build'
        )

    def test_should_parse_flit(tmp_path: Path):
        (tmp_path / 'pyproject.toml').write_text(
            '\n'.join(
                ['[build-system]', 'build-backend = "flit_core.buildapi"']
            )
        )
        assert parsers.PyProjectParser().parse(tmp_path) == BuildStep(
            cmd='flit build'
        )

    def test_should_parse_poetry(tmp_path: Path):
        (tmp_path / 'pyproject.toml').write_text(
            '\n'.join(
                ['[build-system]', 'build-backend = "poetry.core.masonry.api"']
            )
        )
        assert parsers.PyProjectParser().parse(tmp_path) == BuildStep(
            cmd='poetry build'
        )

    def test_should_not_build_if_not_build_system_configured(tmp_path: Path):
        (tmp_path / 'pyproject.toml').write_text(
            '\n'.join(['[tool.poetry]', 'name = "some project"'])
        )
        assert parsers.PyProjectParser().parse(tmp_path) is None


def describe_package_json_parser():
    @pytest.fixture
    def with_package_json(tmp_path: Path):
        (tmp_path / 'package.json').write_text(
            json.dumps(
                {
                    'scripts': {
                        'test': 'some test command',
                        'build': 'some build command',
                    }
                }
            )
        )
        yield tmp_path

    def test_should_should_parse_target_section(with_package_json: Path):
        assert parsers.PackageJsonParser('test').parse(
            with_package_json
        ) == BuildStep(cmd='npm run test')

    def test_should_not_parse_if_section_not_present(with_package_json: Path):
        assert (
            parsers.PackageJsonParser('not exist').parse(with_package_json)
            is None
        )


def describe_docker_parser():
    def test_should_parse_with_docker(tmp_path: Path):
        (tmp_path / 'Dockerfile').write_text(
            '\n'.join(
                [
                    'FROM python3.9',
                    'RUN apt-get update && apt-get install htop',
                ]
            )
        )
        assert parsers.DockerParser().parse(tmp_path) == BuildStep(
            cmd='docker buildx build .'
        )

    def test_should_parse_single_docker_file(tmp_path: Path):
        (tmp_path / 'Dockerfile.service').write_text(
            '\n'.join(
                [
                    'FROM python3.9',
                    'RUN apt-get update && apt-get install htop',
                ]
            )
        )
        assert parsers.DockerParser().parse(tmp_path) == BuildStep(
            cmd='docker buildx build -F Dockerfile.service .'
        )

    def test_should_parse_dockerfile_and_other_docker_file(tmp_path: Path):
        (tmp_path / 'Dockerfile.service').write_text(
            '\n'.join(
                [
                    'FROM python3.9',
                    'RUN apt-get update && apt-get install htop',
                ]
            )
        )
        (tmp_path / 'Dockerfile').write_text(
            '\n'.join(
                [
                    'FROM python3.11',
                    'RUN apt-get update && apt-get install htop',
                ]
            )
        )
        assert parsers.DockerParser().parse(tmp_path) == BuildStep(
            cmd='docker buildx build -F Dockerfile .'
        )

    def test_should_skip_if_no_dockerfile(tmp_path: Path):
        assert parsers.DockerParser().parse(tmp_path) is None



from typing import Iterable
import os
from pathlib import Path
from hashlib import md5

from .ignore import FileSkipperInterface, NeverSkip


def hash_directory(
    path: Path, skipper: FileSkipperInterface = NeverSkip()
) -> str:

    h = md5()
    for filepath in sorted(all_files(path, skipper)):
        h.update(filepath.read_bytes())
    return h.hexdigest()


def all_files(root: Path, skipper: FileSkipperInterface) -> Iterable[Path]:
    skippers: dict[str, FileSkipperInterface] = {}
    for dirname, _, filenames in os.walk(str(root)):
        ignore_file = Path(dirname) / '.abuildignore'
        if ignore_file.exists():
            parent_dir = str(Path(dirname).parent)
            skippers.setdefault(
                dirname,
                skippers.get(parent_dir, skipper).update(ignore_file),
            )
        else:
            parent_dir = str(Path(dirname).parent)
            skippers.setdefault(dirname, skippers.get(parent_dir, skipper))

        for filename in filenames:
            if not skippers[dirname].skip(Path(filename)):
                yield Path(os.path.join(dirname, filename))



from pathlib import Path
import yaml
from pydantic import BaseModel


class BuildStep(BaseModel):
    name: str | None = None
    cmd: str
    break_on_error: bool = True

    @property
    def display_name(self) -> str:
        if self.name:
            return self.name
        else:
            return self.cmd


class Component(BaseModel):
    name: str | None = None
    path: Path
    steps: list[BuildStep]

    @property
    def display_name(self) -> str:
        if self.name:
            return self.name
        else:
            return str(self.path)


class Config(BaseModel):
    components: list[Component]
    state_file: Path = Path('.abuild_state')

    @classmethod
    def from_file(cls, path: Path):
        with path.open('r') as f:
            return cls(**yaml.safe_load(f))
